function [x, niters, rtr] = cgsolve(A, b, x, maxiters, rel_tol)
% cgsolve : Solve Ax=b by conjugate gradients using 
% 
% Algorithm 5.2 (Conjugate Gradient)
%
% Given symmetric positive definite sparse matrix A vector b, 
% and an optional starting point x, 
% this routine runs conjugate gradient to solve for x in A*x=b.
% It iterates until the residual norm is reduced by 10^-6,
% or for at most max(100,sqrt(n)) iterations.
%
% Written by:
% -- 
% John L. Weatherwax                2007-04-02
% 
% email: wax@alum.mit.edu
% 
% Please send comments and especially bug reports to the
% above email address.
% 
%-----

n = length(b);

if( nargin<3 ) % an initial guess at the solution if needed.
  x = diag(diag(A)) \ b; 
end
if( nargin<4 ) % the maximum number of iterations
  maxiters = max(100,sqrt(n));  
end
if( nargin<5 ) % the relative tolarance 
  rel_tol = 1.e-6;
end

normb    = norm(b);         % all variables here are indexed by *k*: 
if( normb==0 ) normb = 1.0; end % for the unique cases where the right-hand-side b=0 
r        = A * x - b;       % initial residual 
p        = -r;              % initial conjugate direction
Ap       = A*p;             % initial product of A and p 
rtr      = r'*r;            % initial residual norm squared 
alpha    = rtr / (p' * Ap); % compute alpha_k 

niters = 0;
while( (sqrt(rtr)/normb > rel_tol) && (niters < maxiters) ) 
    niters = niters+1;
    if( 0 && mod(niters,20)==0 )
      fprintf('performing iteration (k)= %10d; rtr= %10.9f\n',niters,rtr); 
    end

    % transform all variables (alpha,x,r,beta,p) from index k to index k+1: 
    % 
    x      = x + alpha * p;   % x_{k+1} 
    r      = r + alpha * Ap;  % r_{k+1}
    rtrold = rtr;
    rtr    = r'*r;            % r_{k+1}' r_{k+1}
    beta   = rtr / rtrold;    % beta_{k+1} 
    p      = -r + beta * p;   % p_{k+1} 
    Ap     = A * p;           % Ap_{k+1} 
    alpha  = rtr / (p' * Ap); % compute alpha_{k+1} for next pass  
	
end;


